
;;-------------------------------------------------------
;; scent-forager.cl
;; 
;; Build on random-forager.cl and link-activations.cl
;; Borrows from domaccess.cl and ieWithSink.cl to use OLE to connect
;; to Internet Explorer and surf from page to page by following random
;; links. 
;;
;; Just load this file and make a call like
;;    (navigate-scent "computers inkjet printer" 10)
;;
;; Note: currently uses a simple busy-wait loop to lock lisp processing while pages load
;;
;; Peter Pirolli, 4-13-2000
;;-------------------------------------------------------


;;  Initialize ACL's OLE environment

(in-package :user)

(eval-when (compile eval)
   (require :ole-dev))

(eval-when (compile load eval)
   (require :ole))

(ole:require-client-interfaces 
 IConnectionPoint 
 IConnectionPointContainer)

(ole:start-ole)

;;--------------------------------------------------------
;; Global vars


(defvar *history* nil) ;; stores a list of objects representing pages
                       ;; visited (in reverse order)

(defvar *term-db-file* 
  "c:\\program files\\acl50\\Scent\\terms.mdb") ;; Term database file

(defvar *term-db* nil) ;; stores the term database object


(defvar *start-url* "http://www.dmoz.org")  ;; where IE will start

(defvar *current-frame-handle* nil)

;;-------------------------------------------------------------------------------------


;; COM stuff - defines CLSID for IE and IID (interface ID's) for
;; IConnectionPoint, IConnectionPointContainer, IUnknown, and 
;; DWebBrowserEvents2

(defparameter ie-classid
 (ole:unique-guid "{0002DF01-0000-0000-C000-000000000046}"))

(defparameter cp-iid
   (ole:unique-guid "{b196b286-bab4-101a-b69c-00aa00341d07}"))

(defparameter cpc-iid
   (ole:unique-guid "{b196b284-bab4-101a-b69c-00aa00341d07}"))

(defparameter iunk-iid
   (ole:unique-guid "{00000000-0000-0000-c000-000000000046}"))

(defparameter dwbe2-iid
   (ole:unique-guid "{34a715a0-6587-11d0-924a-0020afc7ac4d}"))



;;----------------------------------------------------------------


;;; Explorer objects


(defclass ie ()
  ((ole-object :initarg :ole-object
	       :accessor ole-object)))

(defgeneric print-history-element (elt)
  )

(defmethod print-history-element ((elt ie)) ;; default is to print nothing
   nil)

;; IE objects
(defclass ie-object (ie)
  ())


;; IE events
(defclass ie-event ()
  ())


;; Browser
(defclass browser (ie-object)
  ((document :initarg :document
	     :accessor browser-document)))


;; Document
(defclass document (ie-object)
  ((links :initarg :links
	  :accessor document-links)))

(defmethod print-history-element ((obj document))
   (format t "DOCUMENT~%")
   (mapcar #'print-history-element (document-links obj)))


;; Link
(defclass link (ie-object)
  ((text :initarg :text
	 :accessor link-text)
   (activation :initarg :activation
	       :accessor link-activation)))

(defmethod print-history-element ((obj link))
   (format t "     ~A      ~A~%" (link-text obj) (link-activation obj)))




(defclass Document-Complete (ie-event)
  ((handle :initarg :handle
	   :accessor Document-Complete-handle)
   (url :initarg :url
	:accessor Document-Complete-url)))


(defmethod print-history-element ((obj Document-Complete))
   (format t "DocumentComplete ~A ~A~%" (document-complete-handle obj) (document-complete-url obj)))
	   
;;-------------------------------------------------------------------------------------
;; MS Access object

(defclass term-db ()
  ((ole-object :initarg :ole-object
	       :accessor ole-object)
   (recordset :initarg :recordset
	      :accessor term-db-recordset)
   ))


;;--------------------------------------------------------------------------------------------



(defun navigate-scent (query max-moves)
   (let* ((ietool (ole:ask-for-autotool ie-classid ole:CLSCTX_LOCAL_SERVER)) ;; get a pointer to IE
          (iecpc (ole:query-interface ietool cpc-iid))
          (iecp (ole:find-connection-point iecpc dwbe2-iid))
          (sink (ole:ask-for-autotool iesinkdispatch-classid ole:clsctx_local_server))
          (sinkdisp (ole:query-interface sink idispatch-iid)))
      (start-term-db)
      (ole:advise-connection iecp sinkdisp)
      (setf (ole:auto-getf ietool :visible) t) ;; make IE visible
      ;;(ole:auto-method ietool :gohome) ;; go to the home page
      (ole:auto-method ietool 
        :navigate *start-url*
        0
        ""
        0
        0
        )
      (net-wait ietool) ;; wait for all frames to load
      (do* ((doc (make-instance 'ole:remote-autotool :dispatch (ole:auto-getf ietool :document))
              (make-instance 'ole:remote-autotool :dispatch (ole:auto-getf ietool :document)))
            (links (make-instance 'ole:remote-autotool :dispatch (ole:auto-getf doc :links))
              (make-instance 'ole:remote-autotool :dispatch
			     (ole:auto-getf doc :links)))
            (link-elts (make-link-elts links) (make-link-elts links))
            (elt (highest-scent-link query link-elts) (highest-scent-link query link-elts))
            (url (ole:auto-getf (ole-object elt) :href)
              (ole:auto-getf (ole-object elt) :href))
            (moves max-moves (1- moves)))
           ((zerop moves)                   ;; when done quite IE, release it, and shut down OLE
            (ole:auto-method ietool :quit)
            (ole:release ietool)
            (ole:stop-ole))
	(push (make-instance 'document
			     :ole-object doc
			     :links link-elts)
	      *history*)
         (ole:auto-method ietool :navigate
           url   ; url
           0       ; flags
           ""      ; target frame name
           0       ; postdata
           0       ; extra headers
           
           )
         (net-wait ietool)
         )
      )
   )

(defun make-link-elts (links)
  (do ((i 0 (1+ i)) 
       (nlinks (ole:auto-getf links :length))
       (elts)) 
      ((equal i nlinks) (reverse elts))
    (push (make-instance 'link
			 :ole-object (make-instance 'ole:remote-autotool 
						    :dispatch (ole:auto-method links :item i)))
	  elts)
    (setf (link-text (car elts)) 
          (ole:auto-getf (ole-object (car elts)) :innertext))))

(defun highest-scent-link (query link-elts)
  (mapcar #'(lambda (x)
	      (setf 
             (link-activation x) 
             (summed-activation query (link-text x))))
    link-elts)
  (car 
   (reverse
    (sort link-elts
	  #'(lambda (x y)
	      (< (link-activation x) (link-activation y)))))))
                     
                     
(defun get-idispatch-handle (ietool)
   (ole:h (ole:query-interface ietool idispatch-iid)))


(defun net-wait (tool)
   (format t "*** STARTING NET-WAIT LOOP")
   (do () 
       ((and (not (ole:auto-getf tool :Busy))
             (eq (get-idispatch-handle tool) *current-frame-handle*)
             *navigate-complete*)
        (setq *current-frame-handle* nil)
        (format t "**** NET-WAIT DONE: ALL FRAMES LOADED~%"))
      )
   )


(defun history ()
   (mapcar #'print-history-element *history*))




;;-------------------------------------------------------------------------------------
;; ieSinkDispatch


(defclass ieSinkDispatch (ole:automaton)
    ()
   (:metaclass ole:automaton-class))

;; The CLSID for this class - I generated this CLSID on my machine using guidgen.exe

(defparameter ieSinkDispatch-classid
  (ole:code-guid #x37bb16e0 #xf605 #x11d3 #x83fb #x00805f830522))

(defparameter idispatch-iid (ole:unique-guid "{00020400-0000-0000-c000-000000000046}"))

(ole:def-automethod StatusTextChange ((thisSink ieSinkDispatch) (dispid 102) &rest args)
  (format t "StatusTextChange~%"))

(ole:def-automethod ProgressChange ((thisSink ieSinkDispatch) (dispid 108) &rest args)
  (format t "ProgressChange ~%"))

(ole:def-automethod CommandStateChange ((thisSink ieSinkDispatch) (dispid 105) &rest args)
  (format t "CommandStateChange command = ~d enable = ~a~%" (car args) (second args)))

(ole:def-automethod DownloadBegin ((thisSink ieSinkDispatch) (dispid 106) &rest args)
  (format t "DownloadBegin~%"))

(ole:def-automethod DownloadComplete ((thisSink ieSinkDispatch) (dispid 104) &rest args)
  (format t "DownloadComplete~%"))

(ole:def-automethod TitleChange ((thisSink ieSinkDispatch) (dispid 113) &rest args)
  (format t "TitleChange to ~a~%" (car args)))

(ole:def-automethod PropertyChange ((thisSink ieSinkDispatch) (dispid 112) &rest args)
  (format t "PropertyChange~%"))

(ole:def-automethod BeforeNavigate2 ((thisSink ieSinkDispatch) (dispid 250) &rest args)
  (format t "BeforenNavigate2 URL = ~a~%" (second args)))

(ole:def-automethod NewWindow2 ((thisSink ieSinkDispatch) (dispid 251) &rest args)
  (setq *navigate-complete* nil)
  (format t "NewWindow2~%"))

(ole:def-automethod NavigateComplete2 ((thisSink ieSinkDispatch) (dispid 252) &rest args)
  (setq *navigate-complete* t)
  (format t "NavigateComplete2 URL = ~a~%" (second args)))

(ole:def-automethod DocumentComplete ((thisSink ieSinkDispatch)
				      (dispid 259) &rest args)
		    (push (make-instance 'Document-Complete
					 :handle (ole:h (car args))
					 :url (second args))
			  *history*)
		    (setq *current-frame-handle* (ole:h (car args)))
		    (format t "DocumentComplete URL = ~a~%" (second args)))

(ole:def-automethod OnQuit ((thisSink ieSinkDispatch) (dispid 253) &rest args)
  (format t "OnQuit~%"))

(ole:def-automethod OnVisible ((thisSink ieSinkDispatch) (dispid 254) &rest args)
  (cond ((eql (car args) t) (format t "OnVisible = True~%")) (t (format t "OnVisible = False~%"))))

(ole:def-automethod OnToolBar ((thisSink ieSinkDispatch) (dispid 255) &rest args)
  (cond ((eql (car args) t) (format t "OnToolBar = True~%")) (t (format t "OnToolBar = False~%"))))

(ole:def-automethod OnMenuBar ((thisSink ieSinkDispatch) (dispid 256) &rest args)
  (cond ((eql (car args) t) (format t "OnMenuBar = True~%")) (t (format t "OnMenuBar = False~%"))))

(ole:def-automethod OnStatusBar ((thisSink ieSinkDispatch) (dispid 257) &rest args)
  (cond ((eql (car args) t) (format t "OnStatusBar = True~%")) (t (format t "OnStatusBar = False~%"))))

(ole:def-automethod OnFullScreen ((thisSink ieSinkDispatch) (dispid 258) &rest args)
  (cond ((eql (car args) t) (format t "OnFullScreen = True~%")) (t (format t "OnFullScreen = False~%"))))

(ole:def-automethod OnTheaterMode ((thisSink ieSinkDispatch) (dispid 260) &rest args)
  (cond ((eql (car args) t) (format t "OnTheaterMode = True~%")) (t (format t "OnTheaterMode = False~%"))))

;; Make the factory for the ieSinkDispatch class.  OLE needs a "factory" to construct instances
;; of an object.  The factory is just an object that creates other objects.  We use make-instance
;; to make the factory, then call ole:register-factory to let OLE know of this factory's existence.
;; Now if we call (ole:ask-for-autotool ieSinkDispatch-classid ole:CLSCTX_LOCAL_SERVER) we'll get
;; an instance of a sink.

(defparameter ieSinkDispatch-factory
    (make-instance 'ole:class-factory
      :product-class 'ieSinkDispatch
      :classid ieSinkDispatch-classid))

(ole:register-factory ieSinkDispatch-factory
                ieSinkDispatch-classid
		      ole:CLSCTX_LOCAL_SERVER
		      ole:REGCLS_MULTIPLEUSE)


;;--------------------------------------------------------------------------------------------
;; COM stuff for Access

;; Get a pointer to Access
(defparameter access-clsid (ole:unique-guid "{73a4c9c1-d68d-11d0-98bf-00a0c90dc8d9}"))


(defun start-term-db ()
  (let* ((atool (ole:ask-for-autotool access-clsid ole:CLSCTX_LOCAL_SERVER))
	 (dbetool (make-instance 'ole:remote-autotool 
				 :dispatch (ole:auto-getf atool :dbengine)))
	 (db (make-instance 'ole:remote-autotool 
			    :dispatch (ole:auto-method dbetool
						       :opendatabase *term-db-file*)))
	 (rst (make-instance 'ole:remote-autotool
			     :dispatch (ole:auto-method db :openrecordset "extended-terms"))))
    (setf (ole:auto-getf rst :index) "Term")
    (setq *term-db*
	  (make-instance 'term-db
			 :ole-object dbetool
			 :recordset rst))))

(defun frequency (term)
  (let ((rst (term-db-recordset *term-db*)))
   (ole:auto-method rst :seek "=" term)
   (if (ole:auto-getf rst :nomatch)
      0
     (ole:auto-getf (make-instance 'ole:remote-autotool 
				   :dispatch (ole:auto-getf rst :fields 1)) :value))))


;;;---------------------------------------------------------------------------------------
;; Activation computation


;;; Takes two strings, one representing the query and the other representing link text,
;;; and computes the summed activation spread to the query from the link
;;; e.g., (summed-activation "dog food" "pet store") ==> 56.309223

(defun summed-activation (query link)
   (let ((total-activation 0.0))
      (when (and (stringp query) (stringp link))
         (let* ((query-terms (parse-terms query))
                (link-terms (parse-terms link))
                (base-activation-terms (intersection query-terms link-terms))
                (associated-activation-terms (make-pairs query-terms link-terms))
                )
            (dolist (bt base-activation-terms) ;; sum up the base-level activations
               (setq total-activation (+ total-activation (base-strength bt))))
            (dolist (at associated-activation-terms) ;; sum up the inter-association activations
               (setq total-activation (+ total-activation (association-strength (first at) (second at)))))))
      total-activation))




;;;---------------------------------------------------------------------------------------
;;; String parsing/handling utilities

;;; takes a string and parses it up using whitespace as delimiter
(defun parse-terms (input-string)
   (let* ((string (string-downcase input-string)) ;; make sure everything is downcased
          (end (length string)))
      (do* ((start 0 (1+ next-space))
            (next-space (or (position #\space string) end)
              (if (position #\space (subseq string (1+ next-space)))
                 (+ (position #\space (subseq string (1+ next-space))) (1+ next-space))
                 end))
            (terms (list (subseq string start next-space))
              (cons (subseq string start next-space) terms)))
           ((= next-space end) (reverse terms)))))


;;; takes a list of terms1 and a list of terms2 returns a listof all possible pairings 
;;; of their elements
(defun make-pairs (terms1 terms2)
   (let ((result))
      (dolist (w1 terms1)
         (dolist (w2 terms2)
            (push (list w1 w2) result)))
      result))



;;;------------------------------------------------------------------------
;;; Strengths
;;;------------------------------------------------------------------------
;;; See notes 4/2/96 and 5/9/96

;;; see notes of 5/9/96
(defconstant *strength-normalization* (exp 17))


(defconstant *N* 54964137)


(defun base-strength (word)
  (let* ((Pr[i] (/ (frequency word) *N*))
	 (Pr[-i] (- 1 Pr[i])))
    (log (* (/ Pr[i] Pr[-i]) *strength-normalization*))))



(defun association-strength (word1 word2)
   (let ((f[i] (frequency word2))
         (f[j] (frequency word1))
         (f[ji] (frequency (format nil "~AX~A" word2 word1))))
      (if (or (zerop f[i]) (zerop f[j]) (zerop f[ji]))
         0
         (let* ((Pr[i] (/ f[i] *N*))  ;; prob of i
                (Pr[j] (/ f[j] *N*))  ;; prob j
                (Pr[~i] (- 1 Pr[i]))                ;; prob not i
                (Pr[ji] (/ f[ji] *N*))       ;; prob j and i
                (Pr[j!i] (/ Pr[ji] Pr[i]))          ;; prob j conditional on i
                (Pr[j~i] (- Pr[j] Pr[ji]))          ;; prob j and not i
                (Pr[j!~i] (/ Pr[j~i] Pr[~i]))       ;; prob j conditional not i
                )
            (log (* (/ Pr[j!i] Pr[j!~i]) *strength-normalization*))))
      )
   )




